@@ -1,6 +1,6 @@
 /* cursor.c - the cursor type
  *
- * Copyright (C) 2004-2005 Gerhard Hring <gh@ghaering.de>
+ * Copyright (C) 2004-2006 Gerhard Hring <gh@ghaering.de>
  *
  * This file is part of pysqlite.
  *
@@ -24,8 +24,7 @@
 #include "cursor.h"
 #include "module.h"
 #include "util.h"
-#include "microprotocols.h"
-#include "prepare_protocol.h"
+#include "compat.h"
 
 /* used to decide wether to call PyInt_FromLong or PyLong_FromLongLong */
 #define INT32_MIN (-2147483647 - 1)
@@ -33,7 +32,7 @@
 
 PyObject* cursor_iternext(Cursor *self);
 
-static StatementType detect_statement_type(char* statement)
+static StatementKind detect_statement_type(char* statement)
 {
     char buf[20];
     char* src;
@@ -75,8 +74,6 @@ int cursor_init(Cursor* self, PyObject* args, PyObject* kwargs)
 {
     Connection* connection;
 
-    self->statement = NULL;
-
     if (!PyArg_ParseTuple(args, "O!", &ConnectionType, &connection))
     {
         return -1; 
@@ -88,6 +85,9 @@ int cursor_init(Cursor* self, PyObject* args, PyObject* kwargs)
     self->next_row = NULL;
 
     self->row_cast_map = PyList_New(0);
+    if (!self->row_cast_map) {
+        return -1;
+    }
 
     Py_INCREF(Py_None);
     self->description = Py_None;
@@ -98,6 +98,9 @@ int cursor_init(Cursor* self, PyObject* args, PyObject* kwargs)
     self->arraysize = 1;
 
     self->rowcount = PyInt_FromLong(-1L);
+    if (!self->rowcount) {
+        return -1;
+    }
 
     Py_INCREF(Py_None);
     self->row_factory = Py_None;
@@ -113,11 +116,10 @@ void cursor_dealloc(Cursor* self)
 {
     int rc;
 
-    /* Finalize the statement if the user has not closed the cursor */
+    /* Reset the statement if the user has not closed the cursor */
     if (self->statement) {
-        Py_BEGIN_ALLOW_THREADS
-        rc = sqlite3_finalize(self->statement);
-        Py_END_ALLOW_THREADS
+        rc = statement_reset(self->statement);
+        Py_DECREF(self->statement);
     }
 
     Py_XDECREF(self->connection);
@@ -131,7 +133,7 @@ void cursor_dealloc(Cursor* self)
     self->ob_type->tp_free((PyObject*)self);
 }
 
-void build_row_cast_map(Cursor* self)
+int build_row_cast_map(Cursor* self)
 {
     int i;
     const char* type_start = (const char*)-1;
@@ -144,37 +146,47 @@ void build_row_cast_map(Cursor* self)
     PyObject* key;
 
     if (!self->connection->detect_types) {
-        return;
+        return 0;
     }
 
-    Py_DECREF(self->row_cast_map);
+    Py_XDECREF(self->row_cast_map);
     self->row_cast_map = PyList_New(0);
 
-    for (i = 0; i < sqlite3_column_count(self->statement); i++) {
+    for (i = 0; i < sqlite3_column_count(self->statement->st); i++) {
         converter = NULL;
 
         if (self->connection->detect_types | PARSE_COLNAMES) {
-            colname = sqlite3_column_name(self->statement, i);
+            colname = sqlite3_column_name(self->statement->st, i);
+            if (colname) {
+                for (pos = colname; *pos != 0; pos++) {
+                    if (*pos == '[') {
+                        type_start = pos + 1;
+                    } else if (*pos == ']' && type_start != (const char*)-1) {
+                        key = PyString_FromStringAndSize(type_start, pos - type_start);
+                        if (!key) {
+                            /* creating a string failed, but it is too complicated
+                             * to propagate the error here, we just assume there is
+                             * no converter and proceed */
+                            break;
+                        }
 
-            for (pos = colname; *pos != 0; pos++) {
-                if (*pos == '[') {
-                    type_start = pos + 1;
-                } else if (*pos == ']' && type_start != (const char*)-1) {
-                    key = PyString_FromStringAndSize(type_start, pos - type_start);
-                    converter = PyDict_GetItem(converters, key);
-                    Py_DECREF(key);
-                    break;
+                        converter = PyDict_GetItem(converters, key);
+                        Py_DECREF(key);
+                        break;
+                    }
                 }
-
             }
         }
 
         if (!converter && self->connection->detect_types | PARSE_DECLTYPES) {
-            decltype = sqlite3_column_decltype(self->statement, i);
+            decltype = sqlite3_column_decltype(self->statement->st, i);
             if (decltype) {
                 for (pos = decltype;;pos++) {
                     if (*pos == ' ' || *pos == 0) {
                         py_decltype = PyString_FromStringAndSize(decltype, pos - decltype);
+                        if (!py_decltype) {
+                            return -1;
+                        }
                         break;
                     }
                 }
@@ -184,65 +196,33 @@ void build_row_cast_map(Cursor* self)
             }
         }
 
-        if (converter) {
-            PyList_Append(self->row_cast_map, converter);
-        } else {
-            PyList_Append(self->row_cast_map, Py_None);
+        if (!converter) {
+            converter = Py_None;
+        }
+
+        if (PyList_Append(self->row_cast_map, converter) != 0) {
+            if (converter != Py_None) {
+                Py_DECREF(converter);
+            }
+            Py_XDECREF(self->row_cast_map);
+            self->row_cast_map = NULL;
+
+            return -1;
         }
     }
-}
 
-int _bind_parameter(Cursor* self, int pos, PyObject* parameter)
-{
-    int rc = SQLITE_OK;
-    long longval;
-#ifdef HAVE_LONG_LONG
-    PY_LONG_LONG longlongval;
-#endif
-    const char* buffer;
-    char* string;
-    int buflen;
-    PyObject* stringval;
-
-    if (parameter == Py_None) {
-        rc = sqlite3_bind_null(self->statement, pos);
-    } else if (PyInt_Check(parameter)) {
-        longval = PyInt_AsLong(parameter);
-        rc = sqlite3_bind_int64(self->statement, pos, (sqlite_int64)longval);
-#ifdef HAVE_LONG_LONG
-    } else if (PyLong_Check(parameter)) {
-        longlongval = PyLong_AsLongLong(parameter);
-        /* in the overflow error case, longlongval is -1, and an exception is set */
-        rc = sqlite3_bind_int64(self->statement, pos, (sqlite_int64)longlongval);
-#endif
-    } else if (PyFloat_Check(parameter)) {
-        rc = sqlite3_bind_double(self->statement, pos, PyFloat_AsDouble(parameter));
-    } else if (PyBuffer_Check(parameter)) {
-        if (PyObject_AsCharBuffer(parameter, &buffer, &buflen) == 0) {
-            rc = sqlite3_bind_blob(self->statement, pos, buffer, buflen, SQLITE_TRANSIENT);
-        } else {
-            PyErr_SetString(PyExc_ValueError, "could not convert BLOB to buffer");
-            rc = -1;
-        }
-    } else if PyString_Check(parameter) {
-        string = PyString_AsString(parameter);
-        rc = sqlite3_bind_text(self->statement, pos, string, -1, SQLITE_TRANSIENT);
-    } else if PyUnicode_Check(parameter) {
-        stringval = PyUnicode_AsUTF8String(parameter);
-        string = PyString_AsString(stringval);
-        rc = sqlite3_bind_text(self->statement, pos, string, -1, SQLITE_TRANSIENT);
-        Py_DECREF(stringval);
-    } else {
-        rc = -1;
-    }
-
-    return rc;
+    return 0;
 }
 
 PyObject* _build_column_name(const char* colname)
 {
     const char* pos;
 
+    if (!colname) {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+
     for (pos = colname;; pos++) {
         if (*pos == 0 || *pos == ' ') {
             return PyString_FromStringAndSize(colname, pos - colname);
@@ -250,83 +230,30 @@ PyObject* _build_column_name(const char* colname)
     }
 }
 
-typedef enum {
-    LINECOMMENT_1,
-    IN_LINECOMMENT,
-    COMMENTSTART_1,
-    IN_COMMENT,
-    COMMENTEND_1,
-    NORMAL
-} parse_remaining_sql_state;
-
-/*
- * Checks if there is anything left in an SQL string after SQLite compiled it.
- * This is used to check if somebody tried to execute more than one SQL command
- * with one execute()/executemany() command, which the DB-API and we don't
- * allow.
- *
- * Returns 1 if there is more left than should be. 0 if ok.
- */
-int check_remaining_sql(const char* tail)
+PyObject* unicode_from_string(const char* val_str, int optimize)
 {
-    const char* pos = tail;
+    const char* check;
+    int is_ascii = 0;
 
-    parse_remaining_sql_state state = NORMAL;
+    if (optimize) {
+        is_ascii = 1;
 
-    for (;;) {
-        switch (*pos) {
-            case 0:
-                return 0;
-            case '-':
-                if (state == NORMAL) {
-                    state  = LINECOMMENT_1;
-                } else if (state == LINECOMMENT_1) {
-                    state = IN_LINECOMMENT;
-                }
+        check = val_str;
+        while (*check) {
+            if (*check & 0x80) {
+                is_ascii = 0;
                 break;
-            case ' ':
-            case '\t':
-                break;
-            case '\n':
-            case 13:
-                if (state == IN_LINECOMMENT) {
-                    state = NORMAL;
-                }
-                break;
-            case '/':
-                if (state == NORMAL) {
-                    state = COMMENTSTART_1;
-                } else if (state == COMMENTEND_1) {
-                    state = NORMAL;
-                } else if (state == COMMENTSTART_1) {
-                    return 1;
-                }
-                break;
-            case '*':
-                if (state == NORMAL) {
-                    return 1;
-                } else if (state == LINECOMMENT_1) {
-                    return 1;
-                } else if (state == COMMENTSTART_1) {
-                    state = IN_COMMENT;
-                } else if (state == IN_COMMENT) {
-                    state = COMMENTEND_1;
-                }
-                break;
-            default:
-                if (state == COMMENTEND_1) {
-                    state = IN_COMMENT;
-                } else if (state == IN_LINECOMMENT) {
-                } else if (state == IN_COMMENT) {
-                } else {
-                    return 1;
-                }
+            }
+
+            check++;
         }
-
-        pos++;
     }
 
-    return 0;
+    if (is_ascii) {
+        return PyString_FromString(val_str);
+    } else {
+        return PyUnicode_DecodeUTF8(val_str, strlen(val_str), NULL);
+    }
 }
 
 /*
@@ -339,22 +266,26 @@ PyObject* _fetch_one_row(Cursor* self)
 {
     int i, numcols;
     PyObject* row;
-    PyObject* converted_row;
     PyObject* item = NULL;
     int coltype;
-    long long intval;
+    PY_LONG_LONG intval;
     PyObject* converter;
     PyObject* converted;
-    int nbytes;
+    Py_ssize_t nbytes;
     PyObject* buffer;
     void* raw_buffer;
     const char* val_str;
+    char buf[200];
+    const char* colname;
 
     Py_BEGIN_ALLOW_THREADS
-    numcols = sqlite3_data_count(self->statement);
+    numcols = sqlite3_data_count(self->statement->st);
     Py_END_ALLOW_THREADS
 
     row = PyTuple_New(numcols);
+    if (!row) {
+        return NULL;
+    }
 
     for (i = 0; i < numcols; i++) {
         if (self->connection->detect_types) {
@@ -367,12 +298,15 @@ PyObject* _fetch_one_row(Cursor* self)
         }
 
         if (converter != Py_None) {
-            val_str = (const char*)sqlite3_column_text(self->statement, i);
+            val_str = (const char*)sqlite3_column_text(self->statement->st, i);
             if (!val_str) {
                 Py_INCREF(Py_None);
                 converted = Py_None;
             } else {
                 item = PyString_FromString(val_str);
+                if (!item) {
+                    return NULL;
+                }
                 converted = PyObject_CallFunction(converter, "O", item);
                 if (!converted) {
                     /* TODO: have a way to log these errors */
@@ -384,26 +318,45 @@ PyObject* _fetch_one_row(Cursor* self)
             }
         } else {
             Py_BEGIN_ALLOW_THREADS
-            coltype = sqlite3_column_type(self->statement, i);
+            coltype = sqlite3_column_type(self->statement->st, i);
             Py_END_ALLOW_THREADS
             if (coltype == SQLITE_NULL) {
                 Py_INCREF(Py_None);
                 converted = Py_None;
             } else if (coltype == SQLITE_INTEGER) {
-                intval = sqlite3_column_int64(self->statement, i);
+                intval = sqlite3_column_int64(self->statement->st, i);
                 if (intval < INT32_MIN || intval > INT32_MAX) {
                     converted = PyLong_FromLongLong(intval);
                 } else {
                     converted = PyInt_FromLong((long)intval);
                 }
             } else if (coltype == SQLITE_FLOAT) {
-                converted = PyFloat_FromDouble(sqlite3_column_double(self->statement, i));
+                converted = PyFloat_FromDouble(sqlite3_column_double(self->statement->st, i));
             } else if (coltype == SQLITE_TEXT) {
-                val_str = (const char*)sqlite3_column_text(self->statement, i);
-                converted = PyUnicode_DecodeUTF8(val_str, strlen(val_str), NULL);
+                val_str = (const char*)sqlite3_column_text(self->statement->st, i);
+                if ((self->connection->text_factory == (PyObject*)&PyUnicode_Type)
+                    || (self->connection->text_factory == OptimizedUnicode)) {
+
+                    converted = unicode_from_string(val_str,
+                        self->connection->text_factory == OptimizedUnicode ? 1 : 0);
+
+                    if (!converted) {
+                        colname = sqlite3_column_name(self->statement->st, i);
+                        if (colname) {
+                            colname = "<unknown column name>";
+                        }
+                        PyOS_snprintf(buf, sizeof(buf) - 1, "Could not decode to UTF-8 column %s with text %s",
+                                     colname , val_str);
+                        PyErr_SetString(OperationalError, buf);
+                    }
+                } else if (self->connection->text_factory == (PyObject*)&PyString_Type) {
+                    converted = PyString_FromString(val_str);
+                } else {
+                    converted = PyObject_CallFunction(self->connection->text_factory, "s", val_str);
+                }
             } else {
                 /* coltype == SQLITE_BLOB */
-                nbytes = sqlite3_column_bytes(self->statement, i);
+                nbytes = sqlite3_column_bytes(self->statement->st, i);
                 buffer = PyBuffer_New(nbytes);
                 if (!buffer) {
                     break;
@@ -411,7 +364,7 @@ PyObject* _fetch_one_row(Cursor* self)
                 if (PyObject_AsWriteBuffer(buffer, &raw_buffer, &nbytes)) {
                     break;
                 }
-                memcpy(raw_buffer, sqlite3_column_blob(self->statement, i), nbytes);
+                memcpy(raw_buffer, sqlite3_column_blob(self->statement->st, i), nbytes);
                 converted = buffer;
             }
         }
@@ -420,17 +373,11 @@ PyObject* _fetch_one_row(Cursor* self)
     }
 
     if (PyErr_Occurred()) {
-        return NULL;
-    } else {
-        if (self->row_factory != Py_None) {
-            converted_row = PyObject_CallFunction(self->row_factory, "OO", self, row);
-            Py_DECREF(row);
-        } else {
-            converted_row = row;
-        }
-
-        return converted_row;
+        Py_DECREF(row);
+        row = NULL;
     }
+
+    return row;
 }
 
 PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
@@ -441,20 +388,15 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
     PyObject* parameters_list = NULL;
     PyObject* parameters_iter = NULL;
     PyObject* parameters = NULL;
-    int num_params;
-    const char* tail;
     int i;
     int rc;
     PyObject* func_args;
     PyObject* result;
     int numcols;
+    PY_LONG_LONG lastrowid;
     int statement_type;
     PyObject* descriptor;
-    PyObject* current_param;
-    PyObject* adapted;
     PyObject* second_argument = NULL;
-    int num_params_needed;
-    const char* binding_name;
     long rowcount = 0;
 
     if (!check_thread(self->connection) || !check_connection(self->connection)) {
@@ -482,8 +424,7 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
         } else {
             /* sequence */
             parameters_iter = PyObject_GetIter(second_argument);
-            if (PyErr_Occurred())
-            {
+            if (!parameters_iter) {
                 return NULL;
             }
         }
@@ -505,10 +446,16 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
 
         if (second_argument == NULL) {
             second_argument = PyTuple_New(0);
+            if (!second_argument) {
+                goto error;
+            }
         } else {
             Py_INCREF(second_argument);
         }
-        PyList_Append(parameters_list, second_argument);
+        if (PyList_Append(parameters_list, second_argument) != 0) {
+            Py_DECREF(second_argument);
+            goto error;
+        }
         Py_DECREF(second_argument);
 
         parameters_iter = PyObject_GetIter(parameters_list);
@@ -516,9 +463,7 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
 
     if (self->statement != NULL) {
         /* There is an active statement */
-        Py_BEGIN_ALLOW_THREADS
-        rc = sqlite3_finalize(self->statement);
-        Py_END_ALLOW_THREADS
+        rc = statement_reset(self->statement);
     }
 
     if (PyString_Check(operation)) {
@@ -539,6 +484,9 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
 
     Py_DECREF(self->rowcount);
     self->rowcount = PyInt_FromLong(-1L);
+    if (!self->rowcount) {
+        goto error;
+    }
 
     statement_type = detect_statement_type(operation_cstr);
     if (self->connection->begin_statement) {
@@ -559,9 +507,7 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
                 /* it's a DDL statement or something similar
                    - we better COMMIT first so it works for all cases */
                 if (self->connection->inTransaction) {
-                    func_args = PyTuple_New(0);
-                    result = connection_commit(self->connection, func_args);
-                    Py_DECREF(func_args);
+                    result = connection_commit(self->connection, NULL);
                     if (!result) {
                         goto error;
                     }
@@ -571,127 +517,102 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
             case STATEMENT_SELECT:
                 if (multiple) {
                     PyErr_SetString(ProgrammingError,
-                                "You can only execute SELECT statements in executemany().");
+                                "You cannot execute SELECT statements in executemany().");
                     goto error;
                 }
+                break;
         }
     }
 
-    Py_BEGIN_ALLOW_THREADS
-    rc = sqlite3_prepare(self->connection->db,
-                         operation_cstr,
-                         0,
-                         &self->statement,
-                         &tail);
-    Py_END_ALLOW_THREADS
-    if (rc != SQLITE_OK) {
-        _seterror(self->connection->db);
+    func_args = PyTuple_New(1);
+    if (!func_args) {
+        goto error;
+    }
+    Py_INCREF(operation);
+    if (PyTuple_SetItem(func_args, 0, operation) != 0) {
         goto error;
     }
 
-    if (check_remaining_sql(tail)) {
-        PyErr_SetString(Warning, "You can only execute one statement at a time.");
+    if (self->statement) {
+        (void)statement_reset(self->statement);
+        Py_DECREF(self->statement);
+    }
+
+    self->statement = (Statement*)cache_get(self->connection->statement_cache, func_args);
+    Py_DECREF(func_args);
+
+    if (!self->statement) {
         goto error;
     }
 
-    Py_BEGIN_ALLOW_THREADS
-    num_params_needed = sqlite3_bind_parameter_count(self->statement);
-    Py_END_ALLOW_THREADS
+    if (self->statement->in_use) {
+        Py_DECREF(self->statement);
+        self->statement = PyObject_New(Statement, &StatementType);
+        if (!self->statement) {
+            goto error;
+        }
+        rc = statement_create(self->statement, self->connection, operation);
+        if (rc != SQLITE_OK) {
+            self->statement = 0;
+            goto error;
+        }
+    }
+
+    statement_reset(self->statement);
+    statement_mark_dirty(self->statement);
+
     while (1) {
         parameters = PyIter_Next(parameters_iter);
         if (!parameters) {
             break;
         }
 
-        if (PyDict_Check(parameters)) {
-            /* parameters passed as dictionary */
-            for (i = 1; i <= num_params_needed; i++) {
-                Py_BEGIN_ALLOW_THREADS
-                binding_name = sqlite3_bind_parameter_name(self->statement, i);
-                Py_END_ALLOW_THREADS
-                if (!binding_name) {
-                    PyErr_Format(ProgrammingError, "Binding %d has no name, but you supplied a dictionary (which has only names).", i);
-                    goto error;
-                }
+        statement_mark_dirty(self->statement);
 
-                binding_name++; /* skip first char (the colon) */
-                current_param = PyDict_GetItemString(parameters, binding_name);
-                if (!current_param) {
-                    PyErr_Format(ProgrammingError, "You did not supply a value for binding %d.", i);
-                    goto error;
-                }
-
-                Py_INCREF(current_param);
-                adapted = microprotocols_adapt(current_param, (PyObject*)&SQLitePrepareProtocolType, NULL);
-                if (adapted) {
-                    Py_DECREF(current_param);
-                } else {
-                    PyErr_Clear();
-                    adapted = current_param;
-                }
-
-                rc = _bind_parameter(self, i, adapted);
-                Py_DECREF(adapted);
-
-                if (rc != SQLITE_OK) {
-                    PyErr_Format(InterfaceError, "Error binding parameter :%s - probably unsupported type.", binding_name);
-                    goto error;
-               }
-            }
-        } else {
-            /* parameters passed as sequence */
-            num_params = PySequence_Length(parameters);
-            if (num_params != num_params_needed) {
-                PyErr_Format(ProgrammingError, "Incorrect number of bindings supplied. The current statement uses %d, and there are %d supplied.",
-                             num_params_needed, num_params);
-                goto error;
-            }
-            for (i = 0; i < num_params; i++) {
-                current_param = PySequence_GetItem(parameters, i);
-                adapted = microprotocols_adapt(current_param, (PyObject*)&SQLitePrepareProtocolType, NULL);
-
-                if (adapted) {
-                    Py_DECREF(current_param);
-                } else {
-                    PyErr_Clear();
-                    adapted = current_param;
-                }
-
-                rc = _bind_parameter(self, i + 1, adapted);
-                Py_DECREF(adapted);
-
-                if (rc != SQLITE_OK) {
-                    PyErr_Format(InterfaceError, "Error binding parameter %d - probably unsupported type.", i);
-                    goto error;
-                }
-
-            }
-        }
-
-        build_row_cast_map(self);
-
-        rc = _sqlite_step_with_busyhandler(self->statement, self->connection);
-        if (rc != SQLITE_DONE && rc != SQLITE_ROW) {
-            _seterror(self->connection->db);
+        statement_bind_parameters(self->statement, parameters);
+        if (PyErr_Occurred()) {
             goto error;
         }
 
-        if (rc == SQLITE_ROW) {
-            if (multiple) {
-                PyErr_SetString(ProgrammingError, "executemany() can only execute DML statements.");
+        if (build_row_cast_map(self) != 0) {
+            PyErr_SetString(OperationalError, "Error while building row_cast_map");
+            goto error;
+        }
+
+        rc = _sqlite_step_with_busyhandler(self->statement->st, self->connection);
+        if (rc != SQLITE_DONE && rc != SQLITE_ROW) {
+            rc = statement_reset(self->statement);
+            if (rc == SQLITE_SCHEMA) {
+                rc = statement_recompile(self->statement, parameters);
+                if (rc == SQLITE_OK) {
+                    rc = _sqlite_step_with_busyhandler(self->statement->st, self->connection);
+                } else {
+                    _seterror(self->connection->db);
+                    goto error;
+                }
+            } else {
+                _seterror(self->connection->db);
                 goto error;
             }
+        }
 
+        if (rc == SQLITE_ROW || (rc == SQLITE_DONE && statement_type == STATEMENT_SELECT)) {
             Py_BEGIN_ALLOW_THREADS
-            numcols = sqlite3_data_count(self->statement);
+            numcols = sqlite3_column_count(self->statement->st);
             Py_END_ALLOW_THREADS
 
             if (self->description == Py_None) {
                 Py_DECREF(self->description);
                 self->description = PyTuple_New(numcols);
+                if (!self->description) {
+                    goto error;
+                }
                 for (i = 0; i < numcols; i++) {
                     descriptor = PyTuple_New(7);
-                    PyTuple_SetItem(descriptor, 0, _build_column_name(sqlite3_column_name(self->statement, i)));
+                    if (!descriptor) {
+                        goto error;
+                    }
+                    PyTuple_SetItem(descriptor, 0, _build_column_name(sqlite3_column_name(self->statement->st, i)));
                     Py_INCREF(Py_None); PyTuple_SetItem(descriptor, 1, Py_None);
                     Py_INCREF(Py_None); PyTuple_SetItem(descriptor, 2, Py_None);
                     Py_INCREF(Py_None); PyTuple_SetItem(descriptor, 3, Py_None);
@@ -701,8 +622,19 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
                     PyTuple_SetItem(self->description, i, descriptor);
                 }
             }
+        }
+
+        if (rc == SQLITE_ROW) {
+            if (multiple) {
+                PyErr_SetString(ProgrammingError, "executemany() can only execute DML statements.");
+                goto error;
+            }
 
             self->next_row = _fetch_one_row(self);
+        } else if (rc == SQLITE_DONE && !multiple) {
+            statement_reset(self->statement);
+            Py_DECREF(self->statement);
+            self->statement = 0;
         }
 
         switch (statement_type) {
@@ -720,17 +652,16 @@ PyObject* _query_execute(Cursor* self, int multiple, PyObject* args)
         Py_DECREF(self->lastrowid);
         if (statement_type == STATEMENT_INSERT) {
             Py_BEGIN_ALLOW_THREADS
-            self->lastrowid = PyInt_FromLong((long)sqlite3_last_insert_rowid(self->connection->db));
+            lastrowid = sqlite3_last_insert_rowid(self->connection->db);
             Py_END_ALLOW_THREADS
+            self->lastrowid = PyInt_FromLong((long)lastrowid);
         } else {
             Py_INCREF(Py_None);
             self->lastrowid = Py_None;
         }
 
         if (multiple) {
-            Py_BEGIN_ALLOW_THREADS
-            rc = sqlite3_reset(self->statement);
-            Py_END_ALLOW_THREADS
+            rc = statement_reset(self->statement);
         }
         Py_XDECREF(parameters);
     }
@@ -744,8 +675,8 @@ error:
     if (PyErr_Occurred()) {
         return NULL;
     } else {
-        Py_INCREF(Py_None);
-        return Py_None;
+        Py_INCREF(self);
+        return (PyObject*)self;
     }
 }
 
@@ -766,8 +697,8 @@ PyObject* cursor_executescript(Cursor* self, PyObject* args)
     const char* script_cstr;
     sqlite3_stmt* statement;
     int rc;
-    PyObject* func_args;
     PyObject* result;
+    int statement_completed = 0;
 
     if (!PyArg_ParseTuple(args, "O", &script_obj)) {
         return NULL; 
@@ -781,7 +712,7 @@ PyObject* cursor_executescript(Cursor* self, PyObject* args)
         script_cstr = PyString_AsString(script_obj);
     } else if (PyUnicode_Check(script_obj)) {
         script_str = PyUnicode_AsUTF8String(script_obj);
-        if (!script_obj) {
+        if (!script_str) {
             return NULL;
         }
 
@@ -792,9 +723,7 @@ PyObject* cursor_executescript(Cursor* self, PyObject* args)
     }
 
     /* commit first */
-    func_args = PyTuple_New(0);
-    result = connection_commit(self->connection, func_args);
-    Py_DECREF(func_args);
+    result = connection_commit(self->connection, NULL);
     if (!result) {
         goto error;
     }
@@ -804,10 +733,11 @@ PyObject* cursor_executescript(Cursor* self, PyObject* args)
         if (!sqlite3_complete(script_cstr)) {
             break;
         }
+        statement_completed = 1;
 
         rc = sqlite3_prepare(self->connection->db,
                              script_cstr,
-                             0,
+                             -1,
                              &statement,
                              &script_cstr);
         if (rc != SQLITE_OK) {
@@ -837,8 +767,16 @@ PyObject* cursor_executescript(Cursor* self, PyObject* args)
 error:
     Py_XDECREF(script_str);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    if (!statement_completed) {
+        PyErr_SetString(ProgrammingError, "you did not provide a complete SQL statement");
+    }
+
+    if (PyErr_Occurred()) {
+        return NULL;
+    } else {
+        Py_INCREF(self);
+        return (PyObject*)self;
+    }
 }
 
 PyObject* cursor_getiter(Cursor *self)
@@ -849,6 +787,7 @@ PyObject* cursor_getiter(Cursor *self)
 
 PyObject* cursor_iternext(Cursor *self)
 {
+    PyObject* next_row_tuple;
     PyObject* next_row;
     int rc;
 
@@ -856,23 +795,26 @@ PyObject* cursor_iternext(Cursor *self)
         return NULL;
     }
 
-    if (self->statement == NULL) {
-        PyErr_SetString(ProgrammingError, "no compiled statement - you need to execute() before you can fetch data");
-        return NULL;
-    }
-
     if (!self->next_row) {
-        Py_BEGIN_ALLOW_THREADS
-        (void)sqlite3_finalize(self->statement);
-        Py_END_ALLOW_THREADS
-        self->statement = NULL;
+         if (self->statement) {
+            (void)statement_reset(self->statement);
+            Py_DECREF(self->statement);
+            self->statement = NULL;
+        }
         return NULL;
     }
 
-    next_row = self->next_row;
+    next_row_tuple = self->next_row;
     self->next_row = NULL;
 
-    rc = _sqlite_step_with_busyhandler(self->statement, self->connection);
+    if (self->row_factory != Py_None) {
+        next_row = PyObject_CallFunction(self->row_factory, "OO", self, next_row_tuple);
+        Py_DECREF(next_row_tuple);
+    } else {
+        next_row = next_row_tuple;
+    }
+
+    rc = _sqlite_step_with_busyhandler(self->statement->st, self->connection);
     if (rc != SQLITE_DONE && rc != SQLITE_ROW) {
         Py_DECREF(next_row);
         _seterror(self->connection->db);
@@ -911,9 +853,12 @@ PyObject* cursor_fetchmany(Cursor* self, PyObject* args)
     }
 
     list = PyList_New(0);
+    if (!list) {
+        return NULL;
+    }
 
     /* just make sure we enter the loop */
-    row = (PyObject*)1;
+    row = Py_None;
 
     while (row) {
         row = cursor_iternext(self);
@@ -943,9 +888,12 @@ PyObject* cursor_fetchall(Cursor* self, PyObject* args)
     PyObject* list;
 
     list = PyList_New(0);
+    if (!list) {
+        return NULL;
+    }
 
     /* just make sure we enter the loop */
-    row = (PyObject*)1;
+    row = (PyObject*)Py_None;
 
     while (row) {
         row = cursor_iternext(self);
@@ -972,15 +920,15 @@ PyObject* pysqlite_noop(Connection* self, PyObject* args)
 
 PyObject* cursor_close(Cursor* self, PyObject* args)
 {
-    int rc;
-
     if (!check_thread(self->connection) || !check_connection(self->connection)) {
         return NULL;
     }
 
-    Py_BEGIN_ALLOW_THREADS
-    rc = sqlite3_finalize(self->statement);
-    Py_END_ALLOW_THREADS
+    if (self->statement) {
+        (void)statement_reset(self->statement);
+        Py_DECREF(self->statement);
+        self->statement = 0;
+    }
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -1019,10 +967,13 @@ static struct PyMemberDef cursor_members[] =
     {NULL}
 };
 
+static char cursor_doc[] =
+PyDoc_STR("SQLite database cursor class.");
+
 PyTypeObject CursorType = {
         PyObject_HEAD_INIT(NULL)
         0,                                              /* ob_size */
-        "pysqlite2.dbapi2.Cursor",                      /* tp_name */
+        MODULE_NAME ".Cursor",                          /* tp_name */
         sizeof(Cursor),                                 /* tp_basicsize */
         0,                                              /* tp_itemsize */
         (destructor)cursor_dealloc,                     /* tp_dealloc */
@@ -1041,7 +992,7 @@ PyTypeObject CursorType = {
         0,                                              /* tp_setattro */
         0,                                              /* tp_as_buffer */
         Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_ITER|Py_TPFLAGS_BASETYPE, /* tp_flags */
-        0,                                              /* tp_doc */
+        cursor_doc,                                     /* tp_doc */
         0,                                              /* tp_traverse */
         0,                                              /* tp_clear */
         0,                                              /* tp_richcompare */
@@ -1061,3 +1012,9 @@ PyTypeObject CursorType = {
         0,                                              /* tp_new */
         0                                               /* tp_free */
 };
+
+extern int cursor_setup_types(void)
+{
+    CursorType.tp_new = PyType_GenericNew;
+    return PyType_Ready(&CursorType);
+}
